"""Test that breakpoint() works without -s flag."""

import subprocess
from textwrap import dedent

from snektest import load_fixture, test
from snektest.assertions import assert_eq
from testutils.fixtures import tmp_dir_fixture
from testutils.helpers import create_test_file


@test()
def test_breakpoint_works_without_s_flag() -> None:
    """Test that breakpoint() allows pdb interaction without -s flag.

    This test verifies that when a test contains breakpoint(), the stdin
    proxy detects the read and automatically disables output capture,
    allowing pdb to function properly.
    """
    tmp_dir = load_fixture(tmp_dir_fixture())

    test_file = create_test_file(
        tmp_dir,
        dedent("""
            from snektest import test
            from snektest.assertions import assert_eq

            @test()
            def test_with_breakpoint() -> None:
                x = 1
                breakpoint()
                y = 2
                assert_eq(x + y, 3)
        """),
    )

    result = subprocess.run(
        ["uv", "run", "snektest", str(test_file)],
        input=b"pp locals()\nc\n",
        capture_output=True,
        timeout=5,
        check=False,
    )

    stdout = result.stdout.decode()
    stderr = result.stderr.decode()
    combined_output = stdout + stderr

    expected_output = (
        f"> {test_file}(8)test_with_breakpoint()\n"
        "-> breakpoint()\n"
        "(Pdb) {'x': 1}\n"
        f"(Pdb) {test_file}::test_with_breakpoint ... OK (0.00s)\n"
        "────────────────────────────── 1 passed in 0.00s ───────────────────────────────\n"
    )

    assert_eq(result.returncode, 0)
    assert_eq(combined_output, expected_output)


@test()
def test_breakpoint_can_inspect_variables() -> None:
    """Test that we can inspect variables when breakpoint is hit.

    This verifies that pdb is actually working and we can see variable values.
    """
    tmp_dir = load_fixture(tmp_dir_fixture())

    test_file = create_test_file(
        tmp_dir,
        dedent("""
            from snektest import test
            from snektest.assertions import assert_eq

            @test()
            def test_with_breakpoint() -> None:
                x = 1
                breakpoint()
                y = 2
                assert_eq(x + y, 3)
        """),
    )

    pdb_commands = b"pp x\nc\n"

    result = subprocess.run(
        ["uv", "run", "snektest", str(test_file)],
        input=pdb_commands,
        capture_output=True,
        timeout=5,
        check=False,
    )

    stdout = result.stdout.decode()
    stderr = result.stderr.decode()
    combined_output = stdout + stderr

    expected = dedent(f"""
         > {test_file}(8)test_with_breakpoint()
        -> breakpoint()
        (Pdb) 1
        (Pdb) {test_file}::test_with_breakpoint ... OK (0.00s)
        ────────────────────────────── 1 passed in 0.00s ───────────────────────────────
        """).lstrip()
    assert_eq(combined_output, expected)



from __future__ import annotations

import sys
import warnings
from collections.abc import Callable
from typing import Any, cast

from snektest import assert_eq, assert_in, test
from snektest.output import StdinProxy, capture_output, maybe_capture_output


@test()
def test_capture_output_emits_warnings_and_restores_sys() -> None:
    orig_stdout = sys.stdout
    orig_stderr = sys.stderr
    orig_stdin = sys.stdin

    with capture_output() as (_buf, captured_warnings):
        warnings.warn("hi", stacklevel=1)

    assert_eq(sys.stdout, orig_stdout)
    assert_eq(sys.stderr, orig_stderr)
    assert_eq(sys.stdin, orig_stdin)
    assert_eq(any("hi" in w for w in captured_warnings), True)


@test()
def test_maybe_capture_output_false_branch() -> None:
    with maybe_capture_output(False) as (buf, warning_list):
        assert_eq(buf.getvalue(), "")
        assert_eq(warning_list, [])


@test()
def test_stdin_proxy_covers_proxy_methods_and_properties() -> None:
    calls: list[str] = []

    class DummyStdin:
        closed = False
        encoding = "utf-8"
        mode = "r"
        name = "dummy"

        def __init__(self) -> None:
            self._iter = iter(["line1\n"])

        def __iter__(self) -> Any:
            return self

        def __next__(self) -> str:
            return next(self._iter)

        def read(self, size: int = -1) -> str:
            _ = size
            return "data"

        def readline(self, size: int = -1) -> str:
            _ = size
            return "line\n"

        def readlines(self, hint: int = -1) -> list[str]:
            _ = hint
            return ["a\n", "b\n"]

        def fileno(self) -> int:
            return 0

        def isatty(self) -> bool:
            return False

        some_attr = "x"

    def disable() -> None:
        calls.append("disabled")

    proxy = StdinProxy(cast("Any", DummyStdin()), disable)

    assert_eq(proxy.some_attr, "x")
    assert_eq(proxy.read(), "data")
    assert_eq(proxy.readline(), "line\n")
    assert_eq(proxy.readlines(), ["a\n", "b\n"])

    _ = iter(proxy)
    assert_eq(next(proxy), "line1\n")

    assert_eq(proxy.fileno(), 0)
    assert_eq(proxy.isatty(), False)

    proxy.close()
    assert_eq(proxy.closed, False)
    assert_eq(proxy.encoding, "utf-8")
    assert_eq(proxy.mode, "r")
    assert_eq(proxy.name, "dummy")

    assert_eq(len(calls) >= 1, True)


def _set_breakpoint_hooks(hook: Callable[..., Any]) -> Callable[[], None]:
    original_hook = sys.__breakpointhook__
    sys.__breakpointhook__ = hook
    sys.breakpointhook = hook

    def restore() -> None:
        sys.__breakpointhook__ = original_hook
        sys.breakpointhook = original_hook

    return restore


@test()
def test_breakpoint_custom_hook() -> None:
    called: list[str] = []

    def custom_hook(*args: Any, **kwargs: Any) -> None:
        _ = (args, kwargs)
        called.append("custom")

    sys.breakpointhook = custom_hook
    try:
        with capture_output():
            breakpoint()
    finally:
        sys.breakpointhook = sys.__breakpointhook__

    assert_eq(called, ["custom"])


@test()
def test_breakpoint_args_path() -> None:
    called: list[str] = []

    def dummy_hook(*args: Any, **kwargs: Any) -> None:
        _ = (args, kwargs)
        called.append("args")

    restore = _set_breakpoint_hooks(dummy_hook)
    try:
        with capture_output():
            breakpoint(1)
    finally:
        restore()

    assert_in("args", called)


@test()
def test_breakpoint_missing_frame_path() -> None:
    def dummy_hook(*args: Any, **kwargs: Any) -> None:
        _ = (args, kwargs)

    def frame_provider() -> Any:
        return None

    restore = _set_breakpoint_hooks(dummy_hook)
    try:
        with capture_output(frame_provider=frame_provider):
            breakpoint()
    finally:
        restore()


@test()
def test_breakpoint_inline_pdb_paths() -> None:
    seen: list[str] = []

    class DummyPdb:
        def __init__(self, **kwargs: Any) -> None:
            _ = kwargs

        def message(self, msg: str) -> None:
            seen.append(msg)

        def set_trace(self, frame: Any, commands: Any = None) -> None:
            _ = (frame, commands)

    def noop_hook(*args: Any, **kwargs: Any) -> None:
        _ = (args, kwargs)

    restore = _set_breakpoint_hooks(noop_hook)
    try:
        with capture_output(pdb_factory=DummyPdb):
            sys.settrace(None)
            sys.settrace(None)
            breakpoint(header="HEADER", commands=["c"])
    finally:
        restore()

    assert_eq(seen, ["HEADER"])


@test()
def test_breakpoint_extra_kwargs_path() -> None:
    class DummyPdb:
        def __init__(self, **kwargs: Any) -> None:
            _ = kwargs

        def message(self, msg: str) -> None:
            _ = msg

        def set_trace(self, frame: Any, commands: Any = None) -> None:
            _ = (frame, commands)

    def noop_hook(*args: Any, **kwargs: Any) -> None:
        _ = (args, kwargs)

    restore = _set_breakpoint_hooks(noop_hook)
    try:
        with capture_output(pdb_factory=DummyPdb):
            breakpoint(foo="bar")
    finally:
        restore()



"""Unit tests for Hypothesis integration."""

import inspect
from typing import Any, cast

from hypothesis import strategies as st

from snektest import test
from snektest.assertions import assert_in, assert_raises, assert_true
from snektest.decorators import Marker, test_hypothesis
from snektest.utils import get_test_function_markers, is_test_function


@test()
def test_test_hypothesis_marks_function() -> None:
    @test_hypothesis(st.integers(), mark=Marker.FAST)
    def prop(x: int) -> None:
        _ = x

    assert_true(is_test_function(prop))
    assert_true(get_test_function_markers(prop) == ("fast",))


@test()
def test_test_hypothesis_requires_strategies() -> None:
    with assert_raises(ValueError):
        _ = cast("Any", test_hypothesis)()


@test()
def test_test_hypothesis_rejects_string_markers() -> None:
    with assert_raises(TypeError):
        _ = test_hypothesis(st.just(0), mark="fast")  # pyright: ignore[reportArgumentType]


@test()
def test_test_hypothesis_without_settings_hits_default_settings_path() -> None:
    wrapper = test_hypothesis(st.just(0))

    @wrapper
    def prop(x: int) -> None:
        assert_true(x == 0)

    _ = prop()


@test()
async def test_test_hypothesis_async_schedule_and_on_done_errors() -> None:
    wrapper = test_hypothesis(st.just(0))

    def fake_coroutine_function(*args: Any, **kwargs: Any) -> Any:
        _ = (args, kwargs)
        return None

    fake_marked = inspect.markcoroutinefunction(fake_coroutine_function)
    run_bad = wrapper(fake_marked)

    res = run_bad()
    assert res is not None
    with assert_raises(Exception) as exc_info:
        await res

    assert_in("coroutine", str(exc_info.exception).lower())

    @wrapper
    async def run_raises(_x: int) -> None:
        msg = "boom"
        raise RuntimeError(msg)

    res2 = run_raises()
    assert res2 is not None
    with assert_raises(Exception):
        await res2



"""Meta tests for basic test execution using snektest helper functions."""

from textwrap import dedent

from snektest import load_fixture, test
from snektest.assertions import assert_eq
from testutils.fixtures import tmp_dir_fixture
from testutils.helpers import create_test_file, run_test_subprocess


@test()
async def test_simple_passing_test() -> None:
    """Test that simple passing test works."""
    tmp_dir = load_fixture(tmp_dir_fixture())

    test_file = create_test_file(
        tmp_dir,
        dedent("""
            from snektest import test, assert_true

            @test()
            def test_example() -> None:
                assert_true(True)
        """),
    )

    result = run_test_subprocess(test_file)
    assert_eq(result["passed"], 1)
    assert_eq(result["failed"], 0)
    assert_eq(result["errors"], 0)
    assert_eq(result["returncode"], 0)


@test()
async def test_simple_failing_test() -> None:
    """Test that failures are detected."""
    tmp_dir = load_fixture(tmp_dir_fixture())

    test_file = create_test_file(
        tmp_dir,
        dedent("""
            from snektest import test, fail

            @test()
            def test_will_fail() -> None:
               fail("This test will fail")
        """),
    )

    result = run_test_subprocess(test_file)
    assert_eq(result["passed"], 0)
    assert_eq(result["failed"], 1)
    assert_eq(result["errors"], 0)
    assert_eq(result["returncode"], 1)


@test()
async def test_multiple_tests() -> None:
    """Test running multiple tests in one file."""
    tmp_dir = load_fixture(tmp_dir_fixture())

    test_file = create_test_file(
        tmp_dir,
        dedent("""
            from snektest import test, assert_true

            @test()
            def test_one() -> None:
                assert_true(True)

            @test()
            def test_two() -> None:
                assert_true(True)

            @test()
            def test_three() -> None:
                assert_true(True)
        """),
    )

    result = run_test_subprocess(test_file)
    assert_eq(result["passed"], 3)
    assert_eq(result["failed"], 0)
    assert_eq(result["errors"], 0)
    assert_eq(result["returncode"], 0)



from types import TracebackType
from typing import Any, Self, TypeIs, cast

from snektest.models import AssertionFailure, BadRequestError


def assert_raises[T](
    *expected_exceptions: type[T],
    msg: str | None = None,
) -> RaisesContex[T]:
    return RaisesContex(*expected_exceptions, msg=msg)


class RaisesContex[T]:
    def __init__(self, *expected_exceptions: type[T], msg: str | None = None) -> None:
        self.msg = msg
        self.expected_exceptions = expected_exceptions
        self._exception: T | None = None
        self.has_exited = False

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> bool:
        self.has_exited = True
        if exc_type is None:
            message = (
                self.msg or "Expected to raise an exception but no exception was raised"
            )
            raise AssertionFailure(
                message,
                actual=None,
                expected=self.expected_exceptions,
            )
        if not isinstance(exc_value, self.expected_exceptions):
            expected_exceptions_display_name = " | ".join(
                e.__name__ for e in self.expected_exceptions
            )
            message = (
                self.msg
                or f"Expected to raise {expected_exceptions_display_name} but raised {type(exc_value).__name__}"
            )
            raise AssertionFailure(
                message,
                actual=exc_value,
                expected=self.expected_exceptions,
            )
        self._exception = exc_value
        return True

    @property
    def exception(self) -> T:
        if not self.has_exited:
            msg = "Exception under assert_raises was accessed before exiting the context manager"
            raise BadRequestError(msg)
        return cast("T", self._exception)


def assert_eq(actual: Any, expected: Any, *, msg: str | None = None) -> None:
    """Assert that actual == expected."""
    if actual != expected:
        message = msg or f"{actual!r} != {expected!r}"
        raise AssertionFailure(
            message,
            actual=actual,
            expected=expected,
            operator="==",
        )


def assert_ne(actual: Any, expected: Any, *, msg: str | None = None) -> None:
    """Assert that actual != expected."""
    if actual == expected:
        message = msg or f"{actual!r} == {expected!r}"
        raise AssertionFailure(
            message,
            actual=actual,
            expected=expected,
            operator="!=",
        )


def assert_true(value: Any, *, msg: str | None = None) -> None:
    """Assert that value is True (identity check, not truthiness)."""
    if value is not True:
        message = msg or f"{value!r} is not True"
        raise AssertionFailure(
            message,
            actual=value,
            expected=True,
            operator="is",
        )


def assert_false(value: Any, *, msg: str | None = None) -> None:
    """Assert that value is False (identity check, not falsiness)."""
    if value is not False:
        message = msg or f"{value!r} is not False"
        raise AssertionFailure(
            message,
            actual=value,
            expected=False,
            operator="is",
        )


def assert_is_none(value: Any, *, msg: str | None = None) -> None:
    """Assert that value is None."""
    if value is not None:
        message = msg or f"{value!r} is not None"
        raise AssertionFailure(
            message,
            actual=value,
            expected=None,
            operator="is",
        )


def assert_is_not_none[T](value: T | None, *, msg: str | None = None) -> TypeIs[T]:
    """Assert that value is not None."""
    if value is None:
        message = msg or "value is None"
        raise AssertionFailure(
            message,
            actual=value,
            expected="not None",
            operator="is not",
        )
    return True


def assert_is(actual: Any, expected: Any, *, msg: str | None = None) -> None:
    """Assert that actual is expected (identity check)."""
    if actual is not expected:
        message = msg or f"{actual!r} is not {expected!r}"
        raise AssertionFailure(
            message,
            actual=actual,
            expected=expected,
            operator="is",
        )


def assert_is_not(actual: Any, expected: Any, *, msg: str | None = None) -> None:
    """Assert that actual is not expected (identity check)."""
    if actual is expected:
        message = msg or f"{actual!r} is {expected!r}"
        raise AssertionFailure(
            message,
            actual=actual,
            expected=expected,
            operator="is not",
        )


def assert_lt(actual: Any, expected: Any, *, msg: str | None = None) -> None:
    """Assert that actual < expected."""
    if not actual < expected:
        message = msg or f"{actual!r} >= {expected!r}"
        raise AssertionFailure(
            message,
            actual=actual,
            expected=expected,
            operator="<",
        )


def assert_gt(actual: Any, expected: Any, *, msg: str | None = None) -> None:
    """Assert that actual > expected."""
    if not actual > expected:
        message = msg or f"{actual!r} <= {expected!r}"
        raise AssertionFailure(
            message,
            actual=actual,
            expected=expected,
            operator=">",
        )


def assert_le(actual: Any, expected: Any, *, msg: str | None = None) -> None:
    """Assert that actual <= expected."""
    if not actual <= expected:
        message = msg or f"{actual!r} > {expected!r}"
        raise AssertionFailure(
            message,
            actual=actual,
            expected=expected,
            operator="<=",
        )


def assert_ge(actual: Any, expected: Any, *, msg: str | None = None) -> None:
    """Assert that actual >= expected."""
    if not actual >= expected:
        message = msg or f"{actual!r} < {expected!r}"
        raise AssertionFailure(
            message,
            actual=actual,
            expected=expected,
            operator=">=",
        )


def assert_in(member: Any, container: Any, *, msg: str | None = None) -> None:
    """Assert that member in container."""
    if member not in container:
        message = msg or f"{member!r} not found in {container!r}"
        raise AssertionFailure(
            message,
            actual=member,
            expected=container,
            operator="in",
        )


def assert_not_in(member: Any, container: Any, *, msg: str | None = None) -> None:
    """Assert that member not in container."""
    if member in container:
        message = msg or f"{member!r} found in {container!r}"
        raise AssertionFailure(
            message,
            actual=member,
            expected=container,
            operator="not in",
        )


def assert_isinstance(
    obj: Any, classinfo: type | tuple[type, ...], *, msg: str | None = None
) -> None:
    """Assert that isinstance(obj, classinfo) is True."""
    if not isinstance(obj, classinfo):
        type_name = (
            classinfo.__name__ if isinstance(classinfo, type) else str(classinfo)
        )
        message = msg or f"{obj!r} is not an instance of {type_name}"
        raise AssertionFailure(
            message,
            actual=type(obj).__name__,
            expected=type_name,
            operator="isinstance",
        )


def assert_not_isinstance(
    obj: Any, classinfo: type | tuple[type, ...], *, msg: str | None = None
) -> None:
    """Assert that isinstance(obj, classinfo) is False."""
    if isinstance(obj, classinfo):
        type_name = (
            classinfo.__name__ if isinstance(classinfo, type) else str(classinfo)
        )
        message = msg or f"{obj!r} is an instance of {type_name}"
        raise AssertionFailure(
            message,
            actual=type(obj).__name__,
            expected=f"not {type_name}",
            operator="not isinstance",
        )


def assert_len(obj: Any, expected_length: int, *, msg: str | None = None) -> None:
    """Assert that len(obj) == expected_length."""
    actual_length = len(obj)
    if actual_length != expected_length:
        message = msg or f"Length {actual_length} != {expected_length}"
        raise AssertionFailure(
            message,
            actual=actual_length,
            expected=expected_length,
            operator="len ==",
        )


def fail(msg: str | None = None) -> None:
    """Unconditionally raise an AssertionFailure.

    Args:
        msg: Optional custom message
    """
    message = msg or "Assertion failed"
    raise AssertionFailure(message)



"""Tests for assertion error formatting and diff presentation."""
# TODO: all these tests should assert that the entire output matches the expected output

from typing import Any

from rich.console import Console

from snektest import assert_eq, assert_raises, test
from snektest.models import AssertionFailure
from snektest.presenter.diff import render_assertion_failure


@test()
def test_list_diff_formatting() -> None:
    """Test that list comparison shows clear diff."""
    with assert_raises(AssertionFailure) as exc_info:
        assert_eq(["foo"], ["bar"])

    error_msg = str(exc_info.exception)
    assert_eq(error_msg, "['foo'] != ['bar']")


@test()
def test_nested_list_diff_formatting() -> None:
    """Test that nested list comparison shows clear diff at the right level."""
    with assert_raises(AssertionFailure) as exc_info:
        assert_eq(
            [1, 2, [3, 4, "foo"]],
            [1, 2, [3, 4, "bar"]],
        )

    error_msg = str(exc_info.exception)
    assert_eq(error_msg, "[1, 2, [3, 4, 'foo']] != [1, 2, [3, 4, 'bar']]")


@test()
def test_dict_diff_formatting() -> None:
    """Test that dict comparison shows clear diff."""
    with assert_raises(AssertionFailure) as exc_info:
        assert_eq(
            {"name": "alice", "age": 30},
            {"name": "bob", "age": 30},
        )

    error_msg = str(exc_info.exception)
    assert_eq(error_msg, "{'name': 'alice', 'age': 30} != {'name': 'bob', 'age': 30}")


@test()
def test_multiline_string_diff_formatting() -> None:
    """Test that multiline string comparison shows line-by-line diff."""
    with assert_raises(AssertionFailure) as exc_info:
        assert_eq(
            "hello\nworld\nfoo",
            "hello\nworld\nbar",
        )

    error_msg = str(exc_info.exception)
    assert_eq(error_msg, "'hello\\nworld\\nfoo' != 'hello\\nworld\\nbar'")


@test()
def test_render_assertion_failure_defensive_paths() -> None:
    console = Console(record=True)

    def ndiff_stub(*args: Any, **kwargs: Any) -> list[str]:
        _ = (args, kwargs)
        return ["xx"]

    render_assertion_failure(
        console,
        AssertionFailure("msg", actual=[1], expected=[2]),
        ndiff_func=ndiff_stub,
    )

    render_assertion_failure(
        console,
        AssertionFailure("msg", actual={"a": 1}, expected={"a": 2}),
        ndiff_func=ndiff_stub,
    )


@test()
def test_render_assertion_failure_rich_diff_paths() -> None:
    console = Console(record=True)

    # List diff: index mismatch branch.
    render_assertion_failure(
        console,
        AssertionFailure("msg", actual=[1, 2], expected=[1, 3]),
    )

    # List diff: length mismatch branches.
    render_assertion_failure(
        console,
        AssertionFailure("msg", actual=[1], expected=[1, 2]),
    )
    render_assertion_failure(
        console,
        AssertionFailure("msg", actual=[1, 2], expected=[1]),
    )

    long_list = list(range(60))

    def ndiff_full(*args: Any, **kwargs: Any) -> list[str]:
        _ = (args, kwargs)
        return ["- x", "+ y", "? z", "  w", "xx"]

    render_assertion_failure(
        console,
        AssertionFailure(
            "msg",
            actual={"a": long_list, "b": 2},
            expected={"a": long_list, "b": 3},
        ),
        ndiff_func=ndiff_full,
    )

    def ndiff_multiline(*args: Any, **kwargs: Any) -> list[str]:
        _ = (args, kwargs)
        return ["+ a", "- b", "? c", "  d"]

    render_assertion_failure(
        console,
        AssertionFailure("msg", actual="hallo\nworld", expected="hello\nworld"),
        ndiff_func=ndiff_multiline,
    )
